// Package sdjournal provides a Go interface to the systemd journal read API by
// wrapping the libsystemd C library.
package sdjournal // import "github.com/docker/docker/daemon/logger/journald/internal/sdjournal"
