package transport // import "github.com/docker/docker/pkg/plugins/transport"

// VersionMimetype is the Content-Type the engine sends to plugins.
//
// For convenience, there is an alias in [github.com/docker/docker/pkg/plugins.VersionMimetype].
const VersionMimetype = "application/vnd.docker.plugins.v1.2+json"
