// Package v1 is deprecated and moved to github.com/moby/docker-image-spec/specs-go
//
// Deprecated: use github.com/moby/docker-image-spec/specs-go instead.
package v1

import "github.com/moby/docker-image-spec/specs-go"

const (
	Version      = specs.Version      // Deprecated: use [specs.Version].
	VersionMajor = specs.VersionMajor // Deprecated: use [specs.VersionMajor].
	VersionMinor = specs.VersionMinor // Deprecated: use [specs.VersionMinor].
)
